@@ -4,11 +4,15 @@ import os
 import sys
 import argparse
 from tempfile import mkdtemp
-from os.path import join
+from os.path import join, basename
 
 from dark.fasta import FastaReads
 from dark.process import Executor
 
+IVAR_FREQUENCY_THRESHOLD_DEFAULT = 0.6
+IVAR_DOCS = (
+    'https://andersen-lab.github.io/ivar/html/manualpage.html#autotoc_md19')
+
 
 def main():
     parser = argparse.ArgumentParser(
@@ -55,7 +59,7 @@ def main():
               'in the VCF file will be used.'))
 
     parser.add_argument(
-        '--dryRun', default=False, action='store_true',
+        '--dryRun', action='store_true',
         help='Do not run commands, just print what would be done.')
 
     parser.add_argument(
@@ -65,16 +69,16 @@ def main():
               'turned off. Requires --bam.'))
 
     parser.add_argument(
-        '--log', default=False, action='store_true',
+        '--log', action='store_true',
         help=('Show a log of commands that were (or would be, if --dryRun is '
               'used) executed.'))
 
     parser.add_argument(
-        '--noClean', default=True, action='store_false', dest='clean',
+        '--noClean', action='store_false', dest='clean',
         help=('Do not remove intermediate files or the temporary directory.'))
 
     parser.add_argument(
-        '--callHaplotypesGATK', default=False, action='store_true',
+        '--callHaplotypesGATK', action='store_true',
         help=('Use GATK to call haplotypes. See '
               'https://gatk.broadinstitute.org for details on GATK.'))
 
@@ -84,17 +88,50 @@ def main():
               'https://github.com/broadinstitute/picard for details on '
               'Picard.'))
 
+    parser.add_argument(
+        '--ivar', action='store_true',
+        help='If given, ivar will be used to call the consensus.')
+
+    parser.add_argument(
+        '--ivarFrequencyThreshold', type=float,
+        help=(f'The frequency threshold used by ivar when calling the '
+              f'consensus. If the frequency of the most-common nucleotide at '
+              f'a site meets this threshold, the nucleotide will be called. '
+              f'Otherwise, an ambiguous nucleotide code will be produced, '
+              f'based on the smallest set of most-frequent nucleotides whose '
+              f'summed frequencies meet the threshold. See {IVAR_DOCS} for '
+              f'more information. If not given, '
+              f'{IVAR_FREQUENCY_THRESHOLD_DEFAULT} is used. Can only be used '
+              f'if --ivar is also specified.'))
+
+    parser.add_argument(
+        '--ivarBedFile',
+        help=('If ivar should trim primers, a BED file of the primer '
+              'positions.'))
+
     args = parser.parse_args()
 
     if not (args.bam or args.vcfFile):
         print('At least one of --bam or --vcfFile must be given.',
               file=sys.stderr)
-        sys.exit(0)
+        sys.exit(1)
 
     if args.maskLowCoverage and not args.bam:
         print('If --maskLowCoverage is used, --bam must be too.',
               file=sys.stderr)
-        sys.exit(0)
+        sys.exit(1)
+
+    if args.ivar and not args.bam:
+        print('If --ivar is used, --bam must be too.', file=sys.stderr)
+        sys.exit(1)
+
+    if args.ivarFrequencyThreshold is not None and not args.ivar:
+        print('If --ivarFrequencyThreshold is used, --ivar must be too.',
+              file=sys.stderr)
+        sys.exit(1)
+
+    if args.ivar and args.ivarFrequencyThreshold is None:
+        args.ivarFrequencyThreshold = IVAR_FREQUENCY_THRESHOLD_DEFAULT
 
     e = Executor(args.dryRun)
 
@@ -116,7 +153,7 @@ def main():
                     print('If you use --callHaplotypesGATK, you must give a '
                           'Picard JAR file with --picardJar or else set '
                           'PICARD_JAR in your environment.', file=sys.stderr)
-                    sys.exit(0)
+                    sys.exit(1)
 
             indexFile = args.reference + '.fai'
             if os.path.exists(indexFile):
@@ -178,26 +215,53 @@ def main():
     else:
         result = e.execute(
             "gunzip -c '%s' | egrep -m 1 '^#CHROM' | cut -f10" % vcfFile)
-        sample = result.stdout.strip()
+        sample = 'SAMPLE-NAME' if args.dryRun else result.stdout.strip()
 
     consensusFile = join(tempdir, 'consensus.fasta')
-    result = e.execute(
-        "bcftools consensus --sample '%s' --iupac-codes %s --fasta-ref "
-        "'%s' '%s' > '%s'" %
-        (sample, maskArg, args.reference, vcfFile, consensusFile))
 
-    consensus = list(FastaReads(consensusFile))[0]
+    if args.ivar:
+        if args.ivarBedFile:
+            tempBamFile = join(tempdir, basename(args.bam) + '-trimmed')
+            result = e.execute(
+                "ivar trim -i %r -b %r -p %r -e" % (
+                    args.bam, args.ivarBedFile, tempBamFile))
+            ivarTempBamFile = tempBamFile + '.bam'
+            sortedIvarTempBamFile = tempBamFile + '-trimmed-sorted.bam'
+            result = e.execute(
+                "samtools sort %r -o %r" % (
+                    ivarTempBamFile, sortedIvarTempBamFile))
+            bamFile = sortedIvarTempBamFile
+        else:
+            bamFile = args.bam
 
-    if args.id is not None:
-        consensus.id = args.id
-    elif args.idLambda is not None:
-        idLambda = eval(args.idLambda)
-        consensus.id = idLambda(consensus.id)
+        ivarConsensusFile = join(tempdir, 'temporary-consensus')
+        result = e.execute(
+            "samtools mpileup -A -Q 0 %r | "
+            "ivar consensus -p %r -q 20 -t %r -m %r" % (
+                bamFile, ivarConsensusFile, args.ivarFrequencyThreshold,
+                args.maskLowCoverage))
 
-    print(consensus.toString('fasta'), end='')
+        result = e.execute(
+            "mv %s %s" % (ivarConsensusFile + '.fa', consensusFile))
 
-    if result.stderr:
-        print(result.stderr, end='', file=sys.stderr)
+    else:
+        result = e.execute(
+            "bcftools consensus --sample '%s' --iupac-codes %s --fasta-ref "
+            "'%s' '%s' > '%s'" %
+            (sample, maskArg, args.reference, vcfFile, consensusFile))
+
+        if result.stderr:
+            print(result.stderr, end='', file=sys.stderr)
+
+    if not args.dryRun:
+        consensus = list(FastaReads(consensusFile))[0]
+        if args.id is not None:
+            consensus.id = args.id
+        elif args.idLambda is not None:
+            idLambda = eval(args.idLambda)
+            consensus.id = idLambda(consensus.id)
+
+        print(consensus.toString('fasta'), end='')
 
     if args.dryRun or args.log:
         print('\n'.join(e.log), file=sys.stderr)
